@@ -19,8 +19,10 @@ along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 import argparse
 import base64
+import contextlib
 import hashlib
 import importlib.util
+import io
 import json
 import logging
 import mimetypes
@@ -30,10 +32,11 @@ import shutil
 import subprocess  # nosec
 import sys
 import time
+import typing  # 'from typing import Optional' conflicts with 'from schema import Optional', below
 import zipfile
 from collections import defaultdict
 from collections.abc import Mapping
-from dataclasses import asdict
+from dataclasses import asdict, dataclass
 from datetime import datetime
 
 # Comment below disabling pylint checks is due to a bug in the CircleCi image with Pylint
@@ -138,6 +141,22 @@ class AnalysisContainsDuplicatesException(Exception):
         super().__init__(self.message)
 
 
+@dataclass
+class TestResultContainer:
+    detection: Detection
+    result: TestResult
+    failed_tests: DefaultDict[str, list]
+    output: str
+
+
+@dataclass
+class TestResultsContainer:
+    """A container for all test results"""
+
+    passed: Dict[str, List[TestResultContainer]]
+    errored: Dict[str, List[TestResultContainer]]
+
+
 def load_module(filename: str) -> Tuple[Any, Any]:
     """Loads the analysis function module from a file.
 
@@ -760,6 +779,9 @@ def test_analysis(args: argparse.Namespace) -> Tuple[int, list]:
     log_type_to_data_model, invalid_data_models = setup_data_models(specs[DATAMODEL])
     invalid_specs.extend(invalid_data_models)
 
+    all_test_results = (
+        None if not bool(args.sort_test_results) else TestResultsContainer(passed={}, errored={})
+    )
     # then, import rules and policies; run tests
     failed_tests, invalid_detection = setup_run_tests(
         log_type_to_data_model,
@@ -768,6 +790,7 @@ def test_analysis(args: argparse.Namespace) -> Tuple[int, list]:
         args.skip_disabled_tests,
         destinations_by_name=destinations_by_name,
         ignore_exception_types=ignore_exception_types,
+        all_test_results=all_test_results,
     )
     invalid_specs.extend(invalid_detection)
 
@@ -778,6 +801,21 @@ def test_analysis(args: argparse.Namespace) -> Tuple[int, list]:
     # cleanup tmp global dir
     cleanup_global_helpers(specs[GLOBAL])
 
+    if all_test_results and (all_test_results.passed or all_test_results.errored):
+        for outcome in ["passed", "errored"]:
+            sorted_results = sorted(getattr(all_test_results, outcome).items())
+            for detection_id, test_result_packages in sorted_results:
+                if test_result_packages:
+                    print(detection_id)
+                for test_result_package in test_result_packages:
+                    if len(test_result_package.output) > 0:
+                        print(test_result_package.output)
+                    _print_test_result(
+                        detection=test_result_package.detection,
+                        test_result=test_result_package.result,
+                        failed_tests=test_result_package.failed_tests,
+                    )
+                    print("")
     print_summary(args.path, len(specs[DETECTION]), failed_tests, invalid_specs)
 
     #  if the classic format was invalid, just exit
@@ -880,6 +918,7 @@ def setup_run_tests(  # pylint: disable=too-many-locals,too-many-arguments
     skip_disabled_tests: bool,
     destinations_by_name: Dict[str, FakeDestination],
     ignore_exception_types: List[Type[Exception]],
+    all_test_results: typing.Optional[TestResultsContainer] = None,
 ) -> Tuple[DefaultDict[str, List[Any]], List[Any]]:
     invalid_specs = []
     failed_tests: DefaultDict[str, list] = defaultdict(list)
@@ -907,7 +946,8 @@ def setup_run_tests(  # pylint: disable=too-many-locals,too-many-arguments
                 )
             )
 
-        print(detection.detection_id)
+        if not all_test_results:
+            print(detection.detection_id)
 
         # if there is a setup exception, no need to run tests
         if detection.setup_exception:
@@ -923,8 +963,11 @@ def setup_run_tests(  # pylint: disable=too-many-locals,too-many-arguments
             minimum_tests,
             destinations_by_name,
             ignore_exception_types,
+            all_test_results,
         )
-        print("")
+
+        if not all_test_results:
+            print("")
     return failed_tests, invalid_specs
 
 
@@ -1136,6 +1179,7 @@ def run_tests(  # pylint: disable=too-many-arguments
     minimum_tests: int,
     destinations_by_name: Dict[str, FakeDestination],
     ignore_exception_types: List[Type[Exception]],
+    all_test_results: typing.Optional[TestResultsContainer],
 ) -> DefaultDict[str, list]:
     if len(analysis.get("Tests", [])) < minimum_tests:
         failed_tests[detection.detection_id].append(
@@ -1156,6 +1200,7 @@ def run_tests(  # pylint: disable=too-many-arguments
         failed_tests,
         destinations_by_name,
         ignore_exception_types,
+        all_test_results,
     )
 
     if minimum_tests > 1 and not (
@@ -1176,8 +1221,12 @@ def _run_tests(  # pylint: disable=too-many-arguments
     failed_tests: DefaultDict[str, list],
     destinations_by_name: Dict[str, FakeDestination],
     ignore_exception_types: List[Type[Exception]],
+    all_test_results: typing.Optional[TestResultsContainer],
 ) -> DefaultDict[str, list]:
+    status_passed = "passed"
+    status_errored = "errored"
     for unit_test in tests:
+        test_output = ""
         try:
             entry = unit_test.get("Resource") or unit_test["Log"]
             log_type = entry.get("p_log_type", "")
@@ -1192,11 +1241,18 @@ def _run_tests(  # pylint: disable=too-many-arguments
             test_case: Mapping = entry
             if detection.detection_type.upper() != TYPE_POLICY.upper():
                 test_case = PantherEvent(entry, analysis_data_models.get(log_type))
-            if mock_methods:
-                with patch.multiple(detection.module, **mock_methods):
+            test_output_buf = io.StringIO()
+            with contextlib.redirect_stdout(test_output_buf), contextlib.redirect_stderr(
+                test_output_buf
+            ):
+                if mock_methods:
+                    with patch.multiple(detection.module, **mock_methods):
+                        result = detection.run(
+                            test_case, {}, destinations_by_name, batch_mode=False
+                        )
+                else:
                     result = detection.run(test_case, {}, destinations_by_name, batch_mode=False)
-            else:
-                result = detection.run(test_case, {}, destinations_by_name, batch_mode=False)
+            test_output = test_output_buf.getvalue()
         except (AttributeError, KeyError) as err:
             logging.warning("AttributeError: {%s}", err)
             logging.debug(str(err), exc_info=err)
@@ -1208,6 +1264,10 @@ def _run_tests(  # pylint: disable=too-many-arguments
             logging.debug(str(err), exc_info=err)
             failed_tests[detection.detection_id].append(unit_test["Name"])
             continue
+        finally:
+            if len(test_output) > 0 and not all_test_results:
+                # not buffering results for later sorting; print any output from tests here:
+                print(test_output)
 
         # print results
         spec = TestSpecification(
@@ -1222,7 +1282,21 @@ def _run_tests(  # pylint: disable=too-many-arguments
             ignore_exception_types=ignore_exception_types
         )
 
-        _print_test_result(detection, test_result, failed_tests)
+        if all_test_results:
+            test_result_str = status_passed if test_result.passed else status_errored
+            stored_test_results = getattr(all_test_results, test_result_str)
+            if test_result.detectionId not in stored_test_results:
+                stored_test_results[test_result.detectionId] = []
+            stored_test_results[test_result.detectionId].append(
+                TestResultContainer(
+                    detection=detection,
+                    result=test_result,
+                    failed_tests=failed_tests,
+                    output=test_output,
+                )
+            )
+        else:
+            _print_test_result(detection, test_result, failed_tests)
 
     return failed_tests
 
@@ -1352,6 +1426,15 @@ def setup_parser() -> argparse.ArgumentParser:
         "help": "A destination name that may be returned by the destinations function. "
         "Repeat the argument to define more than one name.",
     }
+    sort_test_results_name = "--sort-test-results"
+    sort_test_results_arg: Dict[str, Any] = {
+        "action": "store_true",
+        "required": False,
+        "default": False,
+        "dest": "sort_test_results",
+        "help": "Sort test results by whether the test passed or failed (passing tests first), "
+        "then by rule ID",
+    }
 
     # -- root parser
 
@@ -1385,6 +1468,7 @@ def setup_parser() -> argparse.ArgumentParser:
     release_parser.add_argument(skip_test_name, **skip_test_arg)
     release_parser.add_argument(skip_disabled_test_name, **skip_disabled_test_arg)
     release_parser.add_argument(available_destination_name, **available_destination_arg)
+    release_parser.add_argument(sort_test_results_name, **sort_test_results_arg)
     release_parser.set_defaults(func=generate_release_assets)
 
     # -- test command
@@ -1399,6 +1483,7 @@ def setup_parser() -> argparse.ArgumentParser:
     test_parser.add_argument(ignore_files_name, **ignore_files_arg)
     test_parser.add_argument(skip_disabled_test_name, **skip_disabled_test_arg)
     test_parser.add_argument(available_destination_name, **available_destination_arg)
+    test_parser.add_argument(sort_test_results_name, **sort_test_results_arg)
     test_parser.set_defaults(func=test_analysis)
 
     # -- publish command
@@ -1478,6 +1563,7 @@ def setup_parser() -> argparse.ArgumentParser:
     upload_parser.add_argument(ignore_extra_keys_name, **ignore_extra_keys_arg)
     upload_parser.add_argument(ignore_files_name, **ignore_files_arg)
     upload_parser.add_argument(available_destination_name, **available_destination_arg)
+    upload_parser.add_argument(sort_test_results_name, **sort_test_results_arg)
     upload_parser.add_argument(batch_uploads_name, **batch_uploads_arg)
     upload_parser.set_defaults(func=pat_utils.func_with_backend(upload_analysis))
 
@@ -1571,6 +1657,7 @@ def setup_parser() -> argparse.ArgumentParser:
     zip_parser.add_argument(skip_test_name, **skip_test_arg)
     zip_parser.add_argument(skip_disabled_test_name, **skip_disabled_test_arg)
     zip_parser.add_argument(available_destination_name, **available_destination_arg)
+    zip_parser.add_argument(sort_test_results_name, **sort_test_results_arg)
     zip_parser.set_defaults(func=zip_analysis)
 
     # -- check-connection command
